import { generateId, Vector } from "./Util";
import { Quaternion } from "./Quaternion";

export function Box (position, size, rotation) {
    this.position = Vector.set(position);
    this.size = size || Box.size(1, 1, 1);
    this.rotation = Quaternion.set(rotation);
    this.id = generateId(4);
    this.data;
}
Box.set = function (box) {
    return new Box(box.position, box.size, box.rotation);
}
Box.size = function (width = 0, height, length) {
    height = height || width;
    length = length || width;
    return { width, height, length };
}
Box.prototype.setId = function (id) {
    this.id = id;
    return this;
}
Box.prototype.setData = function (data) {
    this.data = data;
    return this;
}
Box.prototype.setPosition = function (x, y, z) {
    this.position = Vector.set(x, y, z);
    return this;
}
Box.prototype.setSize = function (width, height, length) {
    this.size = Box.size(width, height, length);
    return this;
}
Box.prototype.setRotation = function (rotation) {
    this.rotation = rotation;
    return this;
}
Box.prototype.rotateVertex = function (vertex) {
    const rotated = this.rotation.multiply(new Quaternion(vertex.x, vertex.y, vertex.z, 0)).multiply(this.rotation.conjugate());
    return Vector.set(rotated);
}
Box.prototype.getVolume = function () {
    return this.size.width * this.size.height * this.size.length;
}
Box.prototype.scale = function (scale) {
    this.size.width *= scale;
    this.size.height *= scale;
    this.size.length *= scale;
}
Box.prototype.getEdges = function () {
    const halfWidth = this.size.width / 2;
    const halfHeight = this.size.height / 2;
    const halfDepth = this.size.length / 2;

    // Define the eight vertices of the box
    const vertices = [
        { x: -halfWidth, y: -halfHeight, z: -halfDepth },
        { x: halfWidth, y: -halfHeight, z: -halfDepth },
        { x: -halfWidth, y: halfHeight, z: -halfDepth },
        { x: halfWidth, y: halfHeight, z: -halfDepth },
        { x: -halfWidth, y: -halfHeight, z: halfDepth },
        { x: halfWidth, y: -halfHeight, z: halfDepth },
        { x: -halfWidth, y: halfHeight, z: halfDepth },
        { x: halfWidth, y: halfHeight, z: halfDepth },
    ];

    // Rotate all vertices according to the box's rotation
    const rotatedVertices = vertices.map(vertex => {
        return this.rotateVertex(vertex);
    });

    // Convert the rotated vertices to an array of points representing the edges
    const edges = [];
    for (let i = 0; i < rotatedVertices.length; i++) {
        for (let j = i + 1; j < rotatedVertices.length; j++) {
            edges.push({
                start: {
                    x: rotatedVertices[i].x + this.position.x,
                    y: rotatedVertices[i].y + this.position.y,
                    z: rotatedVertices[i].z + this.position.z
                },
                end: {
                    x: rotatedVertices[j].x + this.position.x,
                    y: rotatedVertices[j].y + this.position.y,
                    z: rotatedVertices[j].z + this.position.z
                }
            });
        }
    }
    return edges;
};
Box.prototype.lineCollision = function (line) {
    const { p1, p2 } = line;

    function transformPoint(point, position, rotation) {
        const relative = Vector.sub(point, position);
        const rotated = rotation.conjugate().multiply(new Quaternion(relative.x, relative.y, relative.z, 0)).multiply(rotation);
        return { x: rotated.x, y: rotated.y, z: rotated.z };
    }

    const localP1 = transformPoint(p1, this.position, this.rotation);
    const localP2 = transformPoint(p2, this.position, this.rotation);

    // Box bounds in local space
    const min = { x: -this.size.width / 2, y: -this.size.height / 2, z: -this.size.length / 2 };
    const max = { x: this.size.width / 2, y: this.size.height / 2, z: this.size.length / 2 };

    // Ray-Box Intersection logic for line segment
    function rayBoxIntersection(p1, p2, min, max) {
        const dir = { x: p2.x - p1.x, y: p2.y - p1.y, z: p2.z - p1.z };
        const tMin = { x: (min.x - p1.x) / dir.x, y: (min.y - p1.y) / dir.y, z: (min.z - p1.z) / dir.z };
        const tMax = { x: (max.x - p1.x) / dir.x, y: (max.y - p1.y) / dir.y, z: (max.z - p1.z) / dir.z };

        const t1 = {
            x: Math.min(tMin.x, tMax.x),
            y: Math.min(tMin.y, tMax.y),
            z: Math.min(tMin.z, tMax.z),
        };
        const t2 = {
            x: Math.max(tMin.x, tMax.x),
            y: Math.max(tMin.y, tMax.y),
            z: Math.max(tMin.z, tMax.z),
        };

        const tEntry = Math.max(t1.x, t1.y, t1.z);
        const tExit = Math.min(t2.x, t2.y, t2.z);

        // Check if line intersects the box
        return tEntry <= tExit && tExit >= 0 && tEntry <= 1;
    }

    return rayBoxIntersection(localP1, localP2, min, max);
};
Box.prototype.checkCollision = function (box) {
    const edges1 = this.getEdges();
    const edges2 = box.getEdges();
    for (const edge1 of edges1) {
        for (const edge2 of edges2) {
            if (doEdgesIntersect(edge1, edge2)) {
                return true; // Collision detected
            }
        }
    }
    return false; // No collision detected
}
Box.prototype.intersectsSphere = function (p, r = 0.5) {
    const halfSize = new Vector(this.size.width / 2, this.size.height / 2, this.size.length / 2);

    // Transform sphere center to box's local space
    const sphereCenter = Vector.sub(p, this.position);
    const qConjugate = this.rotation.clone().conjugate();
    const localSphereCenter = qConjugate.multiply(new Quaternion(sphereCenter.x, sphereCenter.y, sphereCenter.z, 0)).multiply(this.rotation);
    const localSpherePos = new Vector(localSphereCenter.x, localSphereCenter.y, localSphereCenter.z);

    // Closest point on the box to the sphere center
    const closestPoint = new Vector(
        Math.max(-halfSize.x, Math.min(halfSize.x, localSpherePos.x)),
        Math.max(-halfSize.y, Math.min(halfSize.y, localSpherePos.y)),
        Math.max(-halfSize.z, Math.min(halfSize.z, localSpherePos.z))
    );

    // Check distance between closest point and sphere center
    const distance = Vector.sub(localSpherePos, closestPoint).getMag();

    return distance < r;
}
Box.prototype.intersectsPoint = function (point) {
    const halfSize = new Vector(this.size.width / 2, this.size.height / 2, this.size.length / 2);

    // Transform point to box's local space
    const localPoint = Vector.sub(point, this.position);
    const qConjugate = this.rotation.clone().conjugate();
    const localPointQuat = qConjugate.multiply(new Quaternion(localPoint.x, localPoint.y, localPoint.z, 0)).multiply(this.rotation);
    const localPointPos = new Vector(localPointQuat.x, localPointQuat.y, localPointQuat.z);

    // Check if the point is within the box's bounds
    return (
        localPointPos.x >= -halfSize.x && localPointPos.x <= halfSize.x &&
        localPointPos.y >= -halfSize.y && localPointPos.y <= halfSize.y &&
        localPointPos.z >= -halfSize.z && localPointPos.z <= halfSize.z
    );
}
Box.prototype.intersectsPointNoRot = function (p) {
    const halfSize = new Vector(this.size.width / 2, this.size.height / 2, this.size.length / 2);
    const boxPosition = this.position;

    const closestPoint = {
        x: Math.max(boxPosition.x - halfSize.x, Math.min(p.x, boxPosition.x + halfSize.x)),
        y: Math.max(boxPosition.y - halfSize.y, Math.min(p.y, boxPosition.y + halfSize.y)),
        z: Math.max(boxPosition.z - halfSize.z, Math.min(p.z, boxPosition.z + halfSize.z))
    };

    const distance = Vector.sub(p, closestPoint).getMag();
    return distance < 0.001;
}
function doEdgesIntersect(edge1, edge2) {
    // Check for intersection between two edges
    // Check if the projections of edges overlap along each axis (x, y, z)
    return (
        do1DProjectionOverlap(edge1.start.x, edge1.end.x, edge2.start.x, edge2.end.x) &&
        do1DProjectionOverlap(edge1.start.y, edge1.end.y, edge2.start.y, edge2.end.y) &&
        do1DProjectionOverlap(edge1.start.z, edge1.end.z, edge2.start.z, edge2.end.z)
    );
}
function do1DProjectionOverlap(min1, max1, min2, max2) {
    return (min1 <= max2 && max1 >= min2) || (min2 <= max1 && max2 >= min1);
}
export default { Box };