
export function SpatialHash(size) {
    this.map = {};
    this.size = size;
}
SpatialHash.prototype.getHash = function (position) {
    const x = Math.floor(position.x / this.size);
    const y = Math.floor(position.y / this.size);
    const z = Math.floor(position.z / this.size);
    return `${x},${y},${z}`;
}
SpatialHash.prototype.insert = function (box) {
    let n = this.getHash(box.position);
    if (typeof this.map[n] !== 'object') this.map[n] = [];
    this.map[n].push(box);
}
SpatialHash.prototype.remove = function (box) {
    for (let n in this.map) {
        for (let i = this.map[n].length - 1; i >= 0; i--) {
            if (this.map[n][i].id == box || this.map[n][i].id == box?.id) this.map[n].splice(i, 1);
        }
    }
};
SpatialHash.prototype.isInside = function (v) {
    return this.map[this.getHash(v)] != undefined;
}
SpatialHash.prototype.lineCollision = function (line) {
    const { p1, p2 } = line;

    // Use DDA to trace the line through the grid and collect all hit boxes
    return traceLineDDA(p1, p2, this.size, (x, y, z, hitBoxes) => {
        const cellKey = `${x},${y},${z}`;

        // Skip empty cells
        if (!this.map[cellKey]) return false;

        // Check each box in the cell and collect those hit by the line
        for (const box of this.map[cellKey]) {
            if (box.lineCollision(line)) {
                hitBoxes.push(box); // Add the box to the hit boxes array
            }
        }

        return false; // Continue traversal
    });
};
SpatialHash.prototype.intersectsSphere = function (v, radius) {
    const hash = this.getHash(v), boxes = [];
    if (typeof this.map[hash] === 'object') {
        this.map[hash].forEach(box => {
            if (box.intersectsSphere(v, radius)) boxes.push(box);
        });
    }
    return boxes;
}
SpatialHash.prototype.intersectsPointNoRot = function (v) {
    const hash = this.getHash(v), boxes = [];
    if (typeof this.map[hash] === 'object') {
        this.map[hash].forEach(box => {
            if (box.intersectsPointNoRot(v)) boxes.push(box);
        });
    }
    return boxes;
}

function traceLineDDA(p1, p2, cellSize, callback) {
    // Start and end points in grid coordinates
    let x1 = Math.floor(p1.x / cellSize);
    let y1 = Math.floor(p1.y / cellSize);
    let z1 = Math.floor(p1.z / cellSize);

    const x2 = Math.floor(p2.x / cellSize);
    const y2 = Math.floor(p2.y / cellSize);
    const z2 = Math.floor(p2.z / cellSize);

    // Deltas
    const dx = Math.abs(x2 - x1);
    const dy = Math.abs(y2 - y1);
    const dz = Math.abs(z2 - z1);

    // Steps
    const sx = x1 < x2 ? 1 : -1;
    const sy = y1 < y2 ? 1 : -1;
    const sz = z1 < z2 ? 1 : -1;

    // Initial errors
    let errX = dx / 2;
    let errY = dy / 2;
    let errZ = dz / 2;

    let step = 0;

    const hitBoxes = []; // To store all boxes that get hit

    while (true) {
        // Call the callback for the current cell, passing it the hitBoxes array
        if (callback(x1, y1, z1, hitBoxes)) return hitBoxes;

        // Break the loop when reaching the target cell
        if (x1 === x2 && y1 === y2 && z1 === z2) break;

        // Update errors and move along the dominant direction
        if (errX >= errY && errX >= errZ) {
            errX -= dy;
            errX -= dz;
            x1 += sx;
        } else if (errY >= errX && errY >= errZ) {
            errY -= dx;
            errY -= dz;
            y1 += sy;
        } else {
            errZ -= dx;
            errZ -= dy;
            z1 += sz;
        }

        if (++step >= 64) break;

        // Adjust errors
        errX += dx;
        errY += dy;
        errZ += dz;
    }

    return hitBoxes; // Return all hit boxes
}

export default { SpatialHash };